/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/param.h>

#include <machine/limits.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "bigint.h"

#define BIGINT_WORDSIZE		(sizeof(uint64_t) * CHAR_BIT)

#define BIGINT_NUM_WORDS(h)	(sizeof((h)->un.v1) / sizeof(uint64_t))
#define DIGIT64(x, i, n)	(((i) < (n)) ? (x)->un.v8[i] : 0)
#define DIGIT32(x, i, u)	(((i) >= 2 * (u)) ? 0 : (x)->un.v4[i])

#define HIGH32(x)		(((x) & 0xffffffff00000000) >> 32)
#define LOW32(x)		((uint32_t)(x) & 0xffffffff)
#define MAX32			0xffffffffUL

#ifndef USE_ARG
#define USE_ARG(x)		/*LINTED*/(void)&(x)
#endif

#ifndef __arraycount
#define	__arraycount(__x)	(sizeof(__x) / sizeof(__x[0]))
#endif

#ifndef howmany
#define	howmany(x, y)		(((x) + ((y) - 1)) / (y))
#endif

#define IS_EVEN(x)		(((x)->un.v1[0] % 2) == 0)

/* return the 32-bit number corresponding to the string buffer */
static uint32_t
getword(const char *str, const size_t len, const size_t base, uint64_t *mult)
{
	size_t	size;
	size_t	i;
	char	buf[8 + 1];

	(void) memset(buf, 0x0, sizeof(buf));
	size = (len > 8) ? 8 : len;
	strlcpy(buf, str, size + 1);
	if (mult) {
		for (*mult = 1, i = 0 ; i < size ; i++, *mult *= base) {
		}
	}
	return (uint32_t)strtoul(buf, NULL, (int)base);
}

/* window sizes - hints from openssl */
static inline size_t
getwindowsize(size_t bits)
{
	return (bits > 671) ? 6 :
		(bits > 239) ? 5 :
		(bits > 79) ? 4 :
		(bits > 23) ? 3 : 1;
}

/* used to make const strings into unsigned values we can switch on */
static unsigned
perlhash(const char *s, size_t len)
{
	unsigned	hash;

	for (hash = 0; len-- > 0 ; ) {
		hash = hash * 33 + *s++;
	}
	return hash + (hash >> 5);
}

/* count the number of zeros at start (LSB) of number */
static int
lsb_zero_count(uint4096_t *a)
{
	size_t	words;
	size_t	i;
	size_t	j;
	int	ret;

	if (a == NULL) {
		return 0;
	}
	words = BIGINT_NUM_WORDS(a);
	for (ret = 0, i = 0 ; i < words && a->un.v1 == 0 ; i++) {
		ret += 8;
	}
	if (i < words) {
		for (j = 0 ; j < 8 && a->un.v1[i] & (1 << (unsigned)j) ; j++) {
			ret += 1;
		}
	}
	return ret;
}

/*****************************************************************************/

/* a word operation -- a word is a 64bit unsigned integer */
int
bigint_word_op(uint4096_t *res, const uint4096_t *a, const char *op, const uint64_t w)
{
	uint4096_t	tmp;
	unsigned	cop;
	size_t		words;
	size_t		i;

	switch(cop = perlhash(op, strlen(op))) {
	case /* := */ 0x7f4:	
	case /* <- */  0x828:
	case /* = */  0x3e:
		if (res == NULL) {
			return 0;
		}
		(void) memset(res, 0x0, sizeof(*res));
		res->un.v8[0] = w;
		return 1;
	case /* + */ 0x2c:
	case /* - */ 0x2e:
	case /* * */ 0x2b:
	case /* / */ 0x30:
	case /* % */ 0x26:
		(void) memset(&tmp, 0x0, sizeof(tmp));
		tmp.un.v8[0] = w;
		return bigint_op(res, a, op, &tmp);
	case /* == */ 0x85a:
	case /* ? */ 0x40:
	case /* != */ 0x4a1:
	case /* > */ 0x3f:
	case /* >= */ 0x87c:
	case /* < */ 0x3d:
	case /* <= */ 0x838:
		if (a == NULL) {
			return 0;
		}
                words = BIGINT_NUM_WORDS(a);
                for (i = words - 1 ; i > 0 ; --i) { 
                        if (a->un.v8[i] != 0) {
                                return 0;
                        }
                }
		switch(cop) {
                case /* == */ 0x85a:
		case /* ? */ 0x40:
			return (a->un.v8[0] == w);
		case /* != */ 0x4a1:
			return (a->un.v8[0] != w);
		case /* > */ 0x3f:
			return (a->un.v8[0] > w);
		case /* >= */ 0x87c:
			return (a->un.v8[0] >= w);
		case /* < */ 0x3d:
			return (a->un.v8[0] < w);
		case /* <= */ 0x838:
			return (a->un.v8[0] <= w);
		default:
			break;
		}
		return 0;
	case /* << */ 0x837:
	case /* >> */ 0x87d:
		return bigint_shift(res, a, op, w);
	default:
		return 0;
	}
}

/*************************************************************************/

/* add */
int
bigint_add(uint4096_t *res, const uint4096_t *a, const uint4096_t *b)
{
	uint4096_t	ans;
	uint64_t	carry;
	uint64_t	lo;
	size_t		words;
	size_t		i;

	if (res == NULL || a == NULL || b == NULL) {
		return 0;
	}
	if (bigint_word_op(NULL, a, "==", 0)) {
		(void) memcpy(&ans, b, sizeof(ans));
	} else if (bigint_word_op(NULL, b, "==", 0)) {
		(void) memcpy(&ans, a, sizeof(ans));
	} else {
		words = BIGINT_NUM_WORDS(a) * 2;
		(void) memset(&ans, 0x0, sizeof(ans));
		for (carry = 0, i = 0 ; i < words ; i++) {
			lo = carry + b->un.v4[i] + a->un.v4[i];
			ans.un.v4[i] = LOW32(lo);
			carry = LOW32(lo >> 32);
		}
	}
	(void) memcpy(res, &ans, sizeof(*res));
	return 1;
}

/* subtract */
int
bigint_sub(uint4096_t *res, const uint4096_t *a, const uint4096_t *b)
{
	uint4096_t	ans;
	uint64_t	carry;
	uint64_t	m;
	size_t		words;
	size_t		i;

	if (res == NULL || a == NULL || b == NULL) {
		return 0;
	}
	if (bigint_word_op(NULL, b, "==", 0)) {
		(void) memcpy(&ans, a, sizeof(ans));
	} else {
		words = BIGINT_NUM_WORDS(a) * 2;
		(void) memset(&ans, 0x0, sizeof(ans));
		for (carry = 0, i = 0 ; i < words ; i++) {
			m = (uint64_t)DIGIT32(a, i, words) - DIGIT32(b, i, words) - carry;
			carry = (m > MAX32);
			ans.un.v4[i] = LOW32(m);
		}
	}
	(void) memcpy(res, &ans, sizeof(*res));
	return 1;
}

/* division/modulo */
int
bigint_divmod(uint4096_t *quot, uint4096_t *rem, const uint4096_t *x, const uint4096_t *y)
{
	uint4096_t	q;
	uint4096_t	r;
	uint4096_t	d;
	size_t		xdigits;
	size_t		ydigits;
	size_t		i;

	if (x == NULL || y == NULL) {
		return 0;
	}
	if (bigint_word_op(NULL, y, "==", 0)) {
		return 0;
	}
	(void) memset(&q, 0x0, sizeof(q));
	if (bigint_word_op(NULL, x, "==", 0) || bigint_cmp(x, y) < 0) {
		(void) memcpy(&r, x, sizeof(r));
	} else {
		xdigits = bigint_numbits(x);
		ydigits = bigint_numbits(y);
		(void) memcpy(&d, y, sizeof(d));
		(void) memcpy(&r, x, sizeof(r));
		bigint_shift(&d, &d, "<<", xdigits - ydigits);
		for (i = xdigits - ydigits ; (ssize_t)i >= 0 ; --i) {
			bigint_shift(&q, &q, "<<", 1);
			if (bigint_cmp(&r, &d) >= 0) {
				bigint_bit_set(&q, 0);
				bigint_sub(&r, &r, &d);
			}
			bigint_shift(&d, &d, ">>", 1);
		}
	}
	if (quot) {
		(void) memcpy(quot, &q, sizeof(*quot));
	}
	if (rem) {
		(void) memcpy(rem, &r, sizeof(*rem));
	}
	return 1;
}

/* modulo */
int
bigint_mod(uint4096_t *r, const uint4096_t *x, const uint4096_t *y)
{
	return (r == NULL) ? 0 : bigint_divmod(NULL, r, x, y);
}

/* divide operation */
int
bigint_div(uint4096_t *q, const uint4096_t *x, const uint4096_t *y)
{
	return (q == NULL) ? 0 : bigint_divmod(q, NULL, x, y);
}

/* prod = a * b */
int
bigint_mul(uint4096_t *prod, const uint4096_t *a, const uint4096_t *b)
{
	uint4096_t	ans;
	uint64_t	carry;
	uint64_t	prodlo;
	uint64_t	prodhi;
	uint64_t	loword;
	uint64_t	adigit;
	size_t		words;
	size_t		i;
	size_t		j;

	if (prod == NULL || a == NULL || b == NULL) {
		return 0;
	}
	(void) memset(&ans, 0x0, sizeof(ans));
	if (!bigint_word_op(NULL, a, "==", 0) &&
	    !bigint_word_op(NULL, b, "==", 0)) {
		words = BIGINT_NUM_WORDS(a) * 2;
		for (j = 0; j < words; j++) {
			adigit = DIGIT32(a, j, words);
			for (carry = 0, i = 0; i < words; i++) {
				prodlo = adigit * DIGIT32(b, i, words);
				prodhi = HIGH32(prodlo);
				prodlo = LOW32(prodlo);
				loword = LOW32(ans.un.v4[i + j]) + carry;
				if ((carry = (loword > MAX32)) == 1) { 
					loword &= MAX32;
				}
				if ((loword += prodlo) > MAX32) {
					carry += 1;
					loword &= MAX32;
				}
				carry += prodhi;
				ans.un.v4[i + j] = LOW32(loword);
			}
			if (carry) {
				ans.un.v4[i + j] += LOW32(carry);
			}
		}
	}
	(void) memcpy(prod, &ans, sizeof(*prod));
	return 1;
}

/* wrapper function for arith functions */
int
bigint_op(uint4096_t *res, const uint4096_t *a, const char *op, const uint4096_t *b)
{
	switch((op) ? perlhash(op, strlen(op)) : 0) {
	case /* '+' */ 0x2c:
		return bigint_add(res, a, b);
	case /* '-' */ 0x2e:
		return bigint_sub(res, a, b);
	case /* '*' */ 0x2b:
		return bigint_mul(res, a, b);
	case /* '/' */ 0x30:
		return bigint_divmod(res, NULL, a, b);
	case /* '%' */ 0x26:
		return bigint_divmod(NULL, res, a, b);
	default:
		return 0;
	}
}

/* compare abs(b) and abs(a) */
int
bigint_cmp(const uint4096_t *a, const uint4096_t *b)
{
	size_t	words;
	size_t	i;

	if (a == NULL || b == NULL) {
		return 0;
	}
	words = BIGINT_NUM_WORDS(a);
	for (i = words - 1 ; (ssize_t)i >= 0 ; i--) {
		if (DIGIT64(a, i, words) < DIGIT64(b, i, words)) {
			return -1;
		} else if (DIGIT64(a, i, words) > DIGIT64(b, i, words)) {
			return 1;
		}
	}
	return 0;
}

/* shift left/right by nshift bits */
int
bigint_shift(uint4096_t *res, const uint4096_t *a, const char *op, const uint64_t nshift)
{
	uint4096_t	ans;
	uint64_t	shift;
	uint64_t	carry;
	uint64_t	mask;
	size_t		words;
	size_t		i;

	words = BIGINT_NUM_WORDS(a);
	if (res == NULL || a == NULL || nshift > sizeof(ans.un.v1) * CHAR_BIT) {
		return 0;
	}
	if ((shift = nshift) > BIGINT_WORDSIZE) {
		shift = nshift / CHAR_BIT;
		memset(&ans, 0x0, sizeof(ans));
		if (strcmp(op, ">>") == 0) {
			(void) memcpy(ans.un.v1, &a->un.v1[shift - 1], sizeof(ans.un.v1) - shift);
		} else {
			(void) memcpy(&ans.un.v1[shift - 1], a->un.v1, sizeof(ans.un.v1) - shift);
		}
		shift = nshift % BIGINT_WORDSIZE;
	} else {
		(void) memcpy(&ans, a, sizeof(ans));
	}
	if (shift == 0) {
		(void) memcpy(res, &ans, sizeof(*res));
	} else {
		switch((op) ? perlhash(op, strlen(op)) : 0) {
		case /* >> */ 0x87d:
			/* right shift >> */
			mask = (((uint64_t)1 << nshift) - 1);
			for (carry = 0, i = words - 1 ; (ssize_t)i >= 0 ; --i) {
				res->un.v8[i] = ((ans.un.v8[i] >> nshift) | carry);
				carry = (ans.un.v8[i] & mask) << (BIGINT_WORDSIZE - nshift);
			}
			break;
		case /* << */ 0x837:
			/* left shift << */
			mask = (((uint64_t)1 << shift) - 1) << (BIGINT_WORDSIZE - shift);
			for (carry = 0, i = 0 ; i < words ; i++) {
				res->un.v8[i] = ((ans.un.v8[i] << shift) | carry);
				carry = (ans.un.v8[i] & mask) >> (BIGINT_WORDSIZE - shift);
			}
			break;
		default:
			return 0;
		}
	}
	return 1;
}
 
/* print to stream */
size_t
bigint_fprintf(FILE *fp, const uint4096_t *m)
{
	size_t	cc;
	char	buf[1024 + 1];

	if (m) {
		cc = bigint_snprintf(buf, sizeof(buf), m);
		fprintf(fp, "%s", buf);
		return cc;
	}
	return 0;
}

/* print to buffer */
size_t
bigint_snprintf(char *buf, const size_t size, const uint4096_t *m)
{
	size_t		 words;
	size_t		 off;
	size_t		 cc;
	size_t		 i;
	int		 leading;

	if (m == NULL || buf == NULL) {
		return 0;
	}
	words = BIGINT_NUM_WORDS(m);
	for (off = 0, leading = 1, i = words - 1 ; (ssize_t)i >= 0 && off < size - 1 ; --i) {
		if (leading && m->un.v8[i] != 0) {
			leading = 0;
			if (m->un.v8[i] & 0x8000000000000000ULL) {
				off = snprintf(buf, size, "00");
			}
			cc = snprintf(&buf[off], size - off, "%" PRIx64, m->un.v8[i]);
		} else if (!leading) {
			cc = snprintf(&buf[off], size - off, "%016" PRIx64, m->un.v8[i]);
		} else {
			cc = 0;
		}
		off += cc;
	}
	if (off == 0) {
		buf[off++] = '0';
		buf[off] = 0x0;
	}
	return off;
}

/* convert a hex ASCII string to a fixed-precision integer - base is 10 or 16 */
int
bigint_str2bigint(uint4096_t *a, const char *str, const size_t len, const size_t base)
{
	uint64_t	mult;
	uint64_t	w;
	size_t		i;

	if (a == NULL || str == NULL || base == 0) {
		return 0;
	}
	(void) memset(a, 0x0, sizeof(*a));
	for (i = 0 ; i < len ; i += 8) {
		w = getword(&str[i], len - i, base, &mult);
		bigint_word_op(a, a, "*", mult);
		bigint_word_op(a, a, "+", w);
	}
	return 1;
}

/* put random values into 'a' */
int
bigint_random(uint4096_t *a, const char *type)
{
	size_t		words;
	size_t		last;
	size_t		size;
	size_t		i;
	int		use_arcrandom;

	if (a == NULL) {
		return 0;
	}
	memset(a, 0x0, sizeof(*a));
	words = BIGINT_NUM_WORDS(a);
	use_arcrandom = (type && strcasecmp(type, "strong") == 0);
	size = (use_arcrandom) ? 4 : sizeof(long);
	last = (size == 4) ? words + words : words;
	for (i = 0 ; i < last ; ) {
		if (use_arcrandom) {
			a->un.v4[i++] = arc4random();
			a->un.v4[i++] = arc4random();
		} else if (size == 4) {
			a->un.v4[i++] = (uint32_t)random();
			a->un.v4[i++] = (uint32_t)random();
		} else {
			a->un.v8[i++] = random();
		}
	}
	return 1;
}

/* convert a fixed precision number to a binary */
size_t
bigint_bigint2bin(const uint4096_t *a, uint8_t *buf, size_t size)
{
	size_t		i;

	if (a == NULL || buf == NULL) {
		return 0;
	}
	for (i = 0 ; i < sizeof(a->un.v1) * CHAR_BIT && i < size ; i++) {
		buf[size - 1 - i] = a->un.v1[i];
	}
	return i;
}

/* convert a binary to a fixed precision number */
size_t
bigint_bin2bigint(uint4096_t *a, const uint8_t *buf, size_t size)
{
	size_t		i;

	if (a == NULL || buf == NULL) {
		return 0;
	}
	for (i = 0 ; i < sizeof(a->un.v1) * CHAR_BIT && i < size ; i++) {
		a->un.v1[i] = buf[size - 1 - i];
	}
	return i;
}

/* find the number of significant bits in 'x' */
size_t
bigint_numbits(const uint4096_t *x)
{
	static const uint32_t	nums[] = { 1, 2, 4, 8, 16, 32, 64, 128 };
	size_t			bytes;
	size_t			i;
	size_t			j;

	if (x == 0) {
		return 0;
	}
	bytes = sizeof(x->un.v1);
	for (i = bytes - 1; (ssize_t)i >= 0 && x->un.v1[i] == 0 ; i--) {
	}
	for (j = 0 ; j < __arraycount(nums) && x->un.v1[i] >= nums[j] ; j++) {
	}
	return ((i + 1) * 8) + (j - 1);
}

/* and the number of bytes */
size_t
bigint_numbytes(const uint4096_t *x)
{
	return (x == NULL) ? 0 : howmany(bigint_numbits(x), CHAR_BIT);
}
        
/* test if a bit is set */
int
bigint_bit_is_set(const uint4096_t *a, size_t bit)
{
	if (bit > sizeof(a->un.v1) * CHAR_BIT || a == NULL) {
		return 0;
	}
	return a->un.v1[bit / CHAR_BIT] & (1 << ((unsigned)bit % CHAR_BIT));
}

/* set a bit */
int
bigint_bit_set(uint4096_t *a, size_t bit)
{
	if (bit > sizeof(a->un.v1) * CHAR_BIT || a == NULL) {
		return 0;
	}
	return a->un.v1[bit / CHAR_BIT] |= (1 << ((unsigned)bit % CHAR_BIT));
}

/* clear a bit */
int
bigint_bit_clear(uint4096_t *a, size_t bit)
{
	if (bit > sizeof(a->un.v1) * CHAR_BIT || a == NULL) {
		return 0;
	}
	return a->un.v1[bit / CHAR_BIT] &= ~(1 << ((unsigned)bit % CHAR_BIT));
}

/* bitwise operation - and, or, xor */
int
bigint_bit_op(uint4096_t *res, const uint4096_t *a, char op, const uint4096_t *b)
{
	size_t	words;
	size_t	i;

	if (a == NULL || res == NULL || b == NULL) {
		return 0;
	}
	words = BIGINT_NUM_WORDS(a);
	for (i = 0 ; i < words ; i++) {
		switch(op) {
		case '&':
			res->un.v8[i] = (a->un.v8[i] & b->un.v8[i]);
			break;
		case '|':
			res->un.v8[i] = (a->un.v8[i] | b->un.v8[i]);
			break;
		case '^':
			res->un.v8[i] = (a->un.v8[i] ^ b->un.v8[i]);
			break;
		case '1':
			res->un.v8[i] = 0xffffffffffffffffULL;
			break;
		case '0':
			res->un.v8[i] = 0ULL;
			break;
		default:
			return 0;
		}
	}
	return 1;
}

/* swap two values */
int
bigint_exchange(uint4096_t *a, uint4096_t *b)
{
	uint4096_t	tmp;

	memcpy(&tmp, a, sizeof(tmp));
	memcpy(a, b, sizeof(*a));
	memcpy(b, &tmp, sizeof(*b));
	return 1;
}

/* return the last 64 bits of the number */
int
bigint_uint64(uint4096_t *a, uint64_t *ret)
{
	if (a == NULL || ret == NULL) {
		return 0;
	}
	*ret = a->un.v8[0];
	return 1;
}

/*************************************************************************/
/* higher-level functions */
/*************************************************************************/

/* find gcd */
int
bigint_gcd(uint4096_t *gcd, const uint4096_t *xorig, const uint4096_t *yorig)
{
	uint4096_t	g;
	uint4096_t	x;
	uint4096_t	y;
	uint4096_t	t;
	uint4096_t	zero;

	if (xorig == NULL || yorig == NULL || gcd == NULL) {
		return 0;
	}
	(void) memcpy(&x, xorig, sizeof(x));
	(void) memcpy(&y, yorig, sizeof(y));
	(void) memset(&zero, 0x0, sizeof(zero));
	bigint_word_op(&g, NULL, ":=", 1);
	while (IS_EVEN(&x) && IS_EVEN(&y)) {
		bigint_shift(&x, &x, ">>", 1);
		bigint_shift(&y, &y, ">>", 1);
		bigint_shift(&g, &g, "<<", 1);
	}
	while (bigint_cmp(&x, &zero) != 0) {
		while (IS_EVEN(&x) && bigint_cmp(&x, &zero) != 0) {
			bigint_shift(&x, &x, ">>", 1);
		}
		while (IS_EVEN(&y) && bigint_cmp(&y, &zero) != 0) {
			bigint_shift(&y, &y, ">>", 1);
		}
		if (bigint_cmp(&x, &y) > 0) {
			bigint_sub(&t, &x, &y);
		} else {
			bigint_sub(&t, &y, &x);
		}
		bigint_shift(&t, &t, ">>", 1);
		if (bigint_cmp(&x, &y) >= 0) {
			(void) memcpy(&x, &t, sizeof(x));
		} else {
			(void) memcpy(&y, &t, sizeof(x));
		}
	}
	bigint_mul(gcd, &g, &y);
	return 1;
}

/* exponentiation */
int
bigint_exp(uint4096_t *res, const uint4096_t *a, const uint4096_t *p)
{
	uint4096_t	ans;
	uint4096_t	v;
	size_t		bits;
	size_t		i;

	if (res == NULL || a == NULL || p == NULL) {
		return 0;
	}
	(void) memcpy(&v, a, sizeof(v));
	bits = bigint_numbits(p);
	if (IS_EVEN(p)) {
		bigint_word_op(&ans, NULL, ":=", 1);
	} else {
		(void) memcpy(&ans, a, sizeof(ans));
	}
	for (i = 1 ; i < bits ; i++) {
		bigint_mul(&v, &v, &v);
		if (bigint_bit_is_set(p, i)) {
			bigint_mul(&ans, &ans, &v);
		}
	}
	(void) memcpy(res, &ans, sizeof(*res));
	return 1;
}

/* inverse via modulo */
int
bigint_mod_inverse(uint4096_t *ret, const uint4096_t *in, const uint4096_t *a, const uint4096_t *n)
{
	uint4096_t	tmp;
	uint4096_t	A;
	uint4096_t	B;
	uint4096_t	X;
	uint4096_t	Y;
	uint4096_t	M;
	uint4096_t	D;
	uint4096_t	T;
	uint4096_t	R;

	if (ret == NULL || in == NULL || a == NULL || n == NULL) {
		return 0;
	}
	(void) memset(&X, 0x0, sizeof(X));
	(void) memset(&Y, 0x0, sizeof(Y));
	(void) memset(&M, 0x0, sizeof(M));
	(void) memset(&D, 0x0, sizeof(D));
	(void) memset(&T, 0x0, sizeof(T));
	(void) memcpy(&R, in, sizeof(R));
	bigint_word_op(&X, NULL, ":=", 1);
	bigint_word_op(&Y, NULL, ":=", 0);
	(void) memcpy(&B, a, sizeof(B));
	(void) memcpy(&A, n, sizeof(A));
	while (!bigint_word_op(NULL, &B, "==", 0)) {
		if (bigint_numbits(&A) == bigint_numbits(&B)) {
			bigint_word_op(&D, NULL, ":=", 1);
			bigint_sub(&M, &A, &B);
		} else if (bigint_numbits(&A) == bigint_numbits(&B) + 1) {
			bigint_shift(&T, &B, "<<", 1);
			if (bigint_cmp(&A, &T) < 0) {
				bigint_word_op(&D, NULL, ":=", 1);
				bigint_sub(&M, &A, &B);
			} else {
				bigint_sub(&M, &A, &T);
				bigint_add(&D, &T, &B);
				if (bigint_cmp(&A, &D) < 0) {
					bigint_word_op(&D, NULL, ":=", 2);
				} else {
					bigint_word_op(&D, NULL, ":=", 3);
					bigint_sub(&M, &M, &B);
				}
			}
		} else {
			bigint_divmod(&D, &M, &A, &B);
		}
		(void) memcpy(&tmp, &A, sizeof(tmp));
		(void) memcpy(&A, &B, sizeof(A));
		(void) memcpy(&B, &M, sizeof(B));
		if (bigint_word_op(NULL, &D, "==", 1) == 0) {
			bigint_add(&tmp, &X, &Y);
		} else {
			if (bigint_word_op(NULL, &D, "==", 2)) {
				bigint_shift(&tmp, &X, "<<", 1);
			} else if (bigint_word_op(NULL, &D, "==", 4)) {
				bigint_shift(&tmp, &X, "<<", 2);
			} else {
				bigint_mul(&tmp, &D, &X);
			}
			bigint_add(&tmp, &tmp, &Y);
		}
		(void) memcpy(&M, &Y, sizeof(M));
		(void) memcpy(&Y, &X, sizeof(Y));
		(void) memcpy(&X, &tmp, sizeof(X));
	}
	if (bigint_word_op(NULL, &A, "==", 1)) {
		if (bigint_cmp(&Y, n) < 0) {
			(void) memcpy(&R, &Y, sizeof(R));
		} else {
			bigint_mod(&R, &Y, n);
		}
	} else {
		fprintf(stderr, "no inverse\n");
	}
	(void) memcpy(ret, &R, sizeof(*ret));
	return 1;
}

/* multiplication via modulo */
int
bigint_mod_mul(uint4096_t *r, const uint4096_t *a, const uint4096_t *b, const uint4096_t *m)
{
	uint4096_t	t;

	if (r == NULL || a == NULL || b == NULL || m == NULL) {
		return 0;
	}
	(void) memset(&t, 0x0, sizeof(t));
	bigint_mul(&t, a, b);
	bigint_mod(r, &t, m);
	return 1;
}

/* exponentiation via modulo */
int
bigint_mod_exp(uint4096_t *r, const uint4096_t *a, const uint4096_t *p, const uint4096_t *m)
{
	uint4096_t	val[32];
	uint4096_t	d;
	ssize_t		winfrom;
	ssize_t		i;
	size_t		windowend;
	size_t		windowsize;
	size_t		wvalue;
	size_t		from;
	size_t		bits;
	size_t		j;

	if (r == NULL || a == NULL || p == NULL || m == NULL) {
		return 0;
	}
	if ((bits = bigint_numbits(p)) == 0) {
		bigint_word_op(r, NULL, ":=", 1);
		return 0;
	}
	(void) memset(&d, 0x0, sizeof(d));
	(void) memset(&val[0], 0x0, sizeof(val[0]));
	bigint_mod(&val[0], a, m);
	if (bigint_word_op(NULL, &val[0], "==", 0)) {
		bigint_word_op(r, NULL, ":=", 0);
		return 1;
	}
	/* set up the window values */
	if ((windowsize = getwindowsize(bits)) > 1) {
		bigint_mod_mul(&d, &val[0], &val[0], m);
		j = ((unsigned)1 << ((unsigned)windowsize - 1));
		for (i = 1 ; i < (ssize_t)j ; i++) {
			(void) memset(&val[i], 0x0, sizeof(val[i]));
			bigint_mod_mul(&val[i], &val[i - 1], &d, m);
		}
	}
	from = 1;
	wvalue = 0;
	winfrom = bits - 1;
	windowend = 0;
	bigint_word_op(r, NULL, ":=", 1);
	for (;;) {
		if (!bigint_bit_is_set(p, (size_t)winfrom)) {
			if (!from) {
				bigint_mod_mul(r, r, r, m);
			}
			if (winfrom == 0) {
				break;
			}
			winfrom -= 1;
			continue;
		}
		j = winfrom;
		wvalue = 1;
		windowend = 0;
		for (i = 1 ; i < (ssize_t)windowsize ; i++) {
			if (winfrom - i < 0) {
				break;
			}
			if (bigint_bit_is_set(p, (size_t)(winfrom - i))) {
				wvalue <<= (i - windowend);
				wvalue |= 1;
				windowend = i;
			}
		}
		j = windowend + 1;
		if (!from) {
			for (i = 0 ; i < (ssize_t)j ; i++) {
				bigint_mod_mul(r, r, r, m);
			}
		}
		bigint_mod_mul(r, r, &val[wvalue >> 1], m);
		winfrom -= windowend + 1;
		wvalue = 0;
		from = 0;
		if (winfrom < 0) {
			break;
		}
	}
	return 1;
}

/* return 1 if 'a' is a prime number (in base 'b') */
int
bigint_is_prime(const uint4096_t *a, const uint4096_t *base)
{
	uint4096_t	n1;
	uint4096_t	tmp;
	uint4096_t	y;
	uint4096_t	sqr;
	uint4096_t	s;
	uint4096_t	r;
	int		bits;
	int		j;

	if (a == NULL || base == NULL) {
		return 0;
	}
	if (bigint_word_op(NULL, base, "<=", 1)) {
		return 0;
	}
	memcpy(&n1, a, sizeof(n1));
	bigint_word_op(&n1, &n1, "-", 1);
	memcpy(&r, &n1, sizeof(r));
	bits = lsb_zero_count(&r);
	bigint_word_op(&s, NULL, ":=", bits);
	bigint_word_op(&tmp, NULL, ":=", 2);
	bigint_exp(&tmp, &tmp, &s);
	bigint_div(&r, &r, &tmp);
	bigint_mod_exp(&y, base, &r, a);
	if (bigint_word_op(NULL, &y, "!=", 1) && bigint_cmp(&y, &n1) != 0) {
		for (j = 1 ; j <= bits - 1 && bigint_cmp(&y, &n1) != 0 ; j++) {
			bigint_mod_mul(&sqr, a, a, &y);
			if (bigint_word_op(NULL, &y, "==", 1)) {
				return 0;
			}
		}
		if (bigint_cmp(&y, &n1) != 0) {
			return 0;
		}
	}
	return 1;
}
