/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef BIGINT_H_
#define BIGINT_H_	20120314

#include <sys/types.h>

#include <inttypes.h>
#include <stdio.h>

/* a fixed-precision integer, 4096 bits wide */
typedef struct uint4096_t {
	union un {
		uint64_t	v8[64];
		uint32_t	v4[128];
		uint8_t		v1[512];
	} un;
} uint4096_t;

/* basic bigint arith functions */
int bigint_add(uint4096_t */*res*/, const uint4096_t */*a*/, const uint4096_t */*b*/);
int bigint_sub(uint4096_t */*res*/, const uint4096_t */*y*/, const uint4096_t */*x*/);
int bigint_mul(uint4096_t */*prod*/, const uint4096_t */*a*/, const uint4096_t */*b*/);
int bigint_divmod(uint4096_t */*q*/, uint4096_t */*r*/, const uint4096_t */*a*/, const uint4096_t */*b*/);
int bigint_div(uint4096_t */*q*/, const uint4096_t */*x*/, const uint4096_t */*y*/);
int bigint_mod(uint4096_t */*r*/, const uint4096_t */*x*/, const uint4096_t */*y*/);

/* all-encompassing arith functions */
int bigint_word_op(uint4096_t */*res*/, const uint4096_t */*a*/, const char */*op*/, const uint64_t /*w*/);
int bigint_op(uint4096_t */*res*/, const uint4096_t */*a*/, const char */*op*/, const uint4096_t */*b*/);

/* string io */
size_t bigint_fprintf(FILE */*fp*/, const uint4096_t */*m*/);
size_t bigint_snprintf(char */*buf*/, const size_t /*size*/, const uint4096_t */*m*/);
int bigint_str2bigint(uint4096_t */*a*/, const char */*str*/, const size_t /*len*/, const size_t /*base*/);

/* comparison */
int bigint_cmp(const uint4096_t */*a*/, const uint4096_t */*b*/);

/* random */
int bigint_random(uint4096_t */*a*/, const char */*type*/);

/* binary io conversion */
size_t bigint_bigint2bin(const uint4096_t */*a*/, uint8_t */*buf*/, const size_t /*size*/);
size_t bigint_bin2bigint(uint4096_t */*a*/, const uint8_t */*buf*/, const size_t /*size*/);
int bigint_uint64(uint4096_t */*a*/, uint64_t */*ret*/);

/* bit operations */
size_t bigint_numbits(const uint4096_t */*x*/);
size_t bigint_numbytes(const uint4096_t */*x*/);
int bigint_bit_is_set(const uint4096_t */*a*/, const size_t /*bit*/);
int bigint_bit_set(uint4096_t */*a*/, const size_t /*bit*/);
int bigint_bit_clear(uint4096_t */*a*/, const size_t /*bit*/);
int bigint_bit_op(uint4096_t */*res*/, const uint4096_t */*a*/, char /*op*/, const uint4096_t */*b*/);
int bigint_shift(uint4096_t */*res*/, const uint4096_t */*a*/, const char */*op*/, const uint64_t /*nshift*/);

/* other */
int bigint_gcd(uint4096_t */*gcd*/, const uint4096_t */*x*/, const uint4096_t */*y*/);
int bigint_exp(uint4096_t */*res*/, const uint4096_t */*a*/, const uint4096_t */*p*/);
int bigint_exchange(uint4096_t */*a*/, uint4096_t */*b*/);
int bigint_is_prime(const uint4096_t */*a*/, const uint4096_t */*base*/);

/* modulo n */
int bigint_mod_inverse(uint4096_t */*ret*/, const uint4096_t */*in*/, const uint4096_t */*a*/, const uint4096_t */*n*/);
int bigint_mod_mul(uint4096_t */*r*/, const uint4096_t */*a*/, const uint4096_t */*b*/, const uint4096_t */*m*/);
int bigint_mod_exp(uint4096_t */*r*/, const uint4096_t */*a*/, const uint4096_t */*p*/, const uint4096_t */*m*/);

#endif
