/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "bn.h"

static const size_t	nbits = 4096;

#ifndef USE_ARG
#define USE_ARG(x)		/*LINTED*/(void)&(x)
#endif

BIGNUM *
BN_new(void)
{
	return calloc(1, sizeof(BIGNUM));
}

void
BN_free(BIGNUM *a)
{
	free(a);
}

void
BN_clear(BIGNUM *a)
{
	(void) memset(a, 0x0, sizeof(*a));
}

void
BN_init(BIGNUM *a)
{
	(void) memset(a, 0x0, sizeof(*a));
}

void
BN_clear_free(BIGNUM *a)
{
	(void) memset(a, 0x0, sizeof(*a));
	free(a);
}

BIGNUM *
BN_copy(BIGNUM *a, const BIGNUM *b)
{
	memcpy(a, b, sizeof(*a));
	return a;
}

BIGNUM *
BN_dup(const BIGNUM *a)
{
	BIGNUM	*b;

	b = BN_new();
	return BN_copy(b, a);
}

BIGNUM *
BN_swap(BIGNUM *a, BIGNUM *b)
{
	bigint_exchange(a, b);
	return b;
}

int
BN_num_bytes(const BIGNUM *a)
{
	return (int)bigint_numbytes(a);
}

int
BN_num_bits(const BIGNUM *a)
{
	return (int)bigint_numbits(a);
}

int
BN_num_bits_word(BN_ULONG w)
{
	uint4096_t	tmp;

	(void) memset(&tmp, 0x0, sizeof(tmp));
	tmp.un.v8[0] = w;
	return (int)bigint_numbits(&tmp);
}

int
BN_add(BIGNUM *r, const BIGNUM *a, const BIGNUM *b)
{
	return bigint_op(r, a, "+", b);
}

int
BN_sub(BIGNUM *r, const BIGNUM *a, const BIGNUM *b)
{
	return bigint_op(r, a, "-", b);
}

int
BN_mul(BIGNUM *r, BIGNUM *a, BIGNUM *b, BN_CTX *ctx)
{
	USE_ARG(ctx);
	return bigint_op(r, a, "*", b);
}

int
BN_sqr(BIGNUM *r, BIGNUM *a, BN_CTX *ctx)
{
	USE_ARG(ctx);
	return bigint_op(r, a, "*", a);
}

int
BN_div(BIGNUM *dv, BIGNUM *rem, const BIGNUM *a, const BIGNUM *d, BN_CTX *ctx)
{
	USE_ARG(ctx);
	return bigint_divmod(dv, rem, a, d);
}

int
BN_mod(BIGNUM *rem, const BIGNUM *a, const BIGNUM *m, BN_CTX *ctx)
{
	USE_ARG(ctx);
	return bigint_divmod(NULL, rem, a, m);
}

int
BN_nnmod(BIGNUM *rem, const BIGNUM *a, const BIGNUM *m, BN_CTX *ctx)
{
	return BN_mod(rem, a, m, ctx);
}

int
BN_mod_add(BIGNUM *ret, BIGNUM *a, BIGNUM *b, const BIGNUM *m, BN_CTX *ctx)
{
	BIGNUM	tmp;

	USE_ARG(ctx);
	bigint_op(&tmp, a, "+", b);
	return bigint_op(ret, &tmp, "%", m);
}

int
BN_mod_sub(BIGNUM *ret, BIGNUM *a, BIGNUM *b, const BIGNUM *m, BN_CTX *ctx)
{
	BIGNUM	tmp;

	USE_ARG(ctx);
	bigint_op(&tmp, a, "-", b);
	return bigint_op(ret, &tmp, "%", m);
}

int
BN_mod_mul(BIGNUM *ret, BIGNUM *a, BIGNUM *b, const BIGNUM *m, BN_CTX *ctx)
{
	BIGNUM	tmp;

	USE_ARG(ctx);
	bigint_op(&tmp, a, "*", b);
	return bigint_op(ret, &tmp, "%", m);
}

int
BN_mod_sqr(BIGNUM *ret, BIGNUM *a, const BIGNUM *m, BN_CTX *ctx)
{
	return BN_mod_mul(ret, a, a, m, ctx);
}

int
BN_exp(BIGNUM *r, BIGNUM *a, BIGNUM *p, BN_CTX *ctx)
{
	USE_ARG(ctx);
	return bigint_exp(r, a, p);
}

int
BN_mod_exp(BIGNUM *r, BIGNUM *a, const BIGNUM *p, const BIGNUM *m, BN_CTX *ctx)
{
	BIGNUM	tmp;

	BN_exp(&tmp, a, __UNCONST(p), ctx);
	return bigint_op(r, &tmp, "%", m);
}

int
BN_gcd(BIGNUM *r, BIGNUM *a, BIGNUM *b, BN_CTX *ctx)
{
	USE_ARG(ctx);
	return bigint_gcd(r, a, b);
}

int
BN_add_word(BIGNUM *a, BN_ULONG w)
{
	return bigint_word_op(a, a, "+", w);
}

int
BN_sub_word(BIGNUM *a, BN_ULONG w)
{
	return bigint_word_op(a, a, "-", w);
}

int
BN_mul_word(BIGNUM *a, BN_ULONG w)
{
	return bigint_word_op(a, a, "*", w);
}

BN_ULONG
BN_div_word(BIGNUM *a, BN_ULONG w)
{
	return bigint_word_op(a, a, "/", w);
}

#if 0
/* XXX */
BN_ULONG
BN_mod_word(const BIGNUM *a, BN_ULONG w)
{
	return bigint_word_op(a, a, "%", w);
}
#endif

int
BN_cmp(BIGNUM *a, BIGNUM *b)
{
	return bigint_cmp(a, b);
}

int
BN_ucmp(BIGNUM *a, BIGNUM *b)
{
	return bigint_cmp(a, b);
}

int
BN_is_zero(BIGNUM *a)
{
	return bigint_word_op(NULL, a, "==", 0);
}

int
BN_is_one(BIGNUM *a)
{
	return bigint_word_op(NULL, a, "==", 1);
}

int
BN_is_word(BIGNUM *a, BN_ULONG w)
{
	return bigint_word_op(NULL, a, "==", w);
}

int
BN_is_odd(BIGNUM *a)
{
	return (a->un.v1[0] % 2) == 1;
}

int
BN_zero(BIGNUM *a)
{
	return bigint_word_op(a, NULL, ":=", 0);
}

int
BN_one(BIGNUM *a)
{
	return bigint_word_op(a, NULL, ":=", 1);
}

const BIGNUM *
BN_value_one(void)
{
	static uint4096_t	one;
	static int		init;

	if (!init) {
		init = 1;
		one.un.v8[0] = 1;
	}
	return &one;
}

int
BN_set_word(BIGNUM *a, unsigned long w)
{
	return bigint_word_op(a, NULL, ":=", w);
}

unsigned long
BN_get_word(BIGNUM *a)
{
	return a->un.v8[0];
}

int
BN_rand(BIGNUM *rnd, int bits, int top, int bottom)
{
	/* XXX */
	USE_ARG(bits);
	USE_ARG(top);
	USE_ARG(bottom);
	return bigint_random(rnd, "strong");
}

int
BN_pseudo_rand(BIGNUM *rnd, int bits, int top, int bottom)
{
	/* XXX */
	USE_ARG(bits);
	USE_ARG(top);
	USE_ARG(bottom);
	return bigint_random(rnd, "weak");
}

int
BN_rand_range(BIGNUM *rnd, BIGNUM *range)
{
	/* XXX */
	USE_ARG(rnd);
	USE_ARG(range);
	return 0;
}

int
BN_pseudo_rand_range(BIGNUM *rnd, BIGNUM *range)
{
	/* XXX */
	USE_ARG(rnd);
	USE_ARG(range);
	return 0;
}

BIGNUM *
BN_generate_prime(BIGNUM *ret, int bits,int safe, BIGNUM *add,
	 BIGNUM *rem, void (*callback)(int, int, void *), void *cb_arg)
{
	/* XXX */
	USE_ARG(ret);
	USE_ARG(bits);
	USE_ARG(safe);
	USE_ARG(add);
	USE_ARG(rem);
	USE_ARG(callback);
	USE_ARG(cb_arg);
	return NULL;
}

int
BN_is_prime(const BIGNUM *p, int nchecks,
	void (*callback)(int, int, void *), BN_CTX *ctx, void *cb_arg)
{
	BIGNUM	base;

	USE_ARG(nchecks);
	USE_ARG(callback);
	USE_ARG(ctx);
	USE_ARG(cb_arg);
	bigint_word_op(&base, NULL, ":=", 4096);
	return bigint_is_prime(p, &base);
}

int
BN_set_bit(BIGNUM *a, int n)
{
	return bigint_bit_set(a, n);
}

int
BN_clear_bit(BIGNUM *a, int n)
{
	return bigint_bit_clear(a, n);
}

int
BN_is_bit_set(const BIGNUM *a, int n)
{
	return bigint_bit_is_set(a, n);
}

int
BN_mask_bits(BIGNUM *a, int n)
{
	/* XXX */
	USE_ARG(a);
	USE_ARG(n);
	return 1;
}

int
BN_lshift(BIGNUM *r, const BIGNUM *a, int n)
{
	return bigint_word_op(r, a, "<<", n);
}

int
BN_lshift1(BIGNUM *r, BIGNUM *a)
{
	return bigint_word_op(r, a, "<<", 1);
}

int
BN_rshift(BIGNUM *r, BIGNUM *a, int n)
{
	return bigint_word_op(r, a, ">>", n);
}

int
BN_rshift1(BIGNUM *r, BIGNUM *a)
{
	return bigint_word_op(r, a, ">>", 1);
}

int
BN_bn2bin(const BIGNUM *a, unsigned char *to)
{
	return (int)bigint_bigint2bin(a, to, 1024);
}

BIGNUM *
BN_bin2bn(const unsigned char *s, int len, BIGNUM *ret)
{
	bigint_bin2bigint(ret, s, len);
	return ret;
}

char *
BN_bn2hex(const BIGNUM *a)
{
	char	*ret;

	if ((ret = calloc(1, 1024)) != NULL) {
		bigint_snprintf(ret, 1024, a);
	}
	return ret;
}

char *
BN_bn2dec(const BIGNUM *a)
{
	/* XXX */
	USE_ARG(a);
	return NULL;
}

int
BN_hex2bn(BIGNUM **a, const char *str)
{
	return bigint_str2bigint(*a, str, strlen(str), 16);
}

int
BN_dec2bn(BIGNUM **a, const char *str)
{
	return bigint_str2bigint(*a, str, strlen(str), 10);
}

int
BN_print(BIO *fp, const BIGNUM *a)
{
	/* XXX */
	USE_ARG(a);
	USE_ARG(fp);
	return 0;
}

int
BN_print_fp(FILE *fp, const BIGNUM *a)
{
	return (int)bigint_fprintf(fp, a);
}

int
BN_bn2mpi(const BIGNUM *a, unsigned char *to)
{
	/* XXX */
	USE_ARG(a);
	USE_ARG(to);
	return 0;
}

BIGNUM *
BN_mpi2bn(unsigned char *s, int len, BIGNUM *ret)
{
	/* XXX */
	USE_ARG(s);
	USE_ARG(len);
	USE_ARG(ret);
	return NULL;
}

BIGNUM *
BN_mod_inverse(BIGNUM *r, BIGNUM *a, const BIGNUM *n, BN_CTX *ctx)
{
	bigint_mod_inverse(r, r, a, n);
	return r;
}
